import waba.fx.*;     // Graphics, Rect
//import waba.io.*;
//import waba.sys.*;    // Vm
import waba.ui.*;
import waba.util.*;   // Vector
import extra.ui.*;    // !!!b. wextras: List (Popup, RelativeContainer, PreferredSize), Title

/**
 * @author <A HREF="mailto:weyer@kagi.com">Steve Weyer</A>
 * Copyright (c) 2001, S. Weyer. All right reserved.
 * @version waba 1.0
 * !!!c 08 Feb 2001 diskList init w/o Vector; move animG init; adjusted speeds
 * !!!b 13 Jan 2001: wextras, move out of bounds bug
 * !!!a 25 Nov 2000
 * <A HREF="http://www.kagi.com/weyer/#Hanoi">Steve's other Hanoi versions</A>
 * <A HREF="http://www.kagi.com/weyer/#Waba">Steve's other Waba apps</A>
 *
 * ! Newton: classheap: 22000
 * ?? how to 'alert'(flash) if illegal move?
 * ?? font metrics for Palm, WinCE?
**/

public class Hanoi extends MainWindow {
  private static final boolean DEBUG = false;

  private static final int COMPHT  = 15;
  private static final int COMPWID = 30;
  private static final int COMPSEP = 2;

  private static final int NUMSTACKS = 3; // !!!b
  private static final int DISKHEIGHT = 10;
  private static final int STACKTOP = 2*COMPHT + COMPSEP;
  private static final int DISKINC = 2;   // animation step (factor of DISKHEIGHT for vertical)
  private int stackBottom;
  private int stackWidth;
  private int numDisks = 3;
  private int maxDiskNum; // calc at startup
  private int moveNum = 0;
  private Timer appTimer = null;
  private int penX;

  private static final int[] POW2 = {
	0x1,	0x2,	0x4,	0x8,
	0x10,	0x20,	0x40,	0x80,
	0x100,	0x200,	0x400,	0x800,
	0x1000,	0x2000,	0x4000,	0x8000,
	0x10000,0x20000,0x40000,0x80000
  };

  private static final String[] SPEEDS = {"slow", "medium", "fast"};
  private static final int[] DELAYS = {500, 50, 1};
  private int delay = DELAYS[DELAYS.length-1]; // default: fast

  // note: 'auto' methods between user and cheat
  private static final String[] METHODS = {"user", "jump", "move", "cheat"};
  private static final int METHUSER = 0;
  private static final int METHJUMP = 1;
  private static final int METHMOVE = 2;
  private static final int METHCHEAT = 3;
  private int moveMeth = METHUSER; // default: user
  private static final String[] STACKNAMES = {"Left", "Middle", "Right"}; // only needed for DEBUG

  private int[] animArgs = {0, 0, 0, 0, 0, DISKHEIGHT, ANIM_NONE, 0, 0};
  private static final int ANIM_FRX = 0;
  private static final int ANIM_FRY   = 1;
  private static final int ANIM_TOX = 2;
  private static final int ANIM_TOY   = 3;
  private static final int ANIM_WID   = 4;
  private static final int ANIM_HT    = 5;
  private static final int ANIM_DIR   = 6;
    private static final int ANIM_NONE    = -1;
    private static final int ANIM_UP      = 0;
    private static final int ANIM_ACROSS  = 1;
    private static final int ANIM_DOWN    = 2;
  private static final int ANIM_VAR   = 7;
  private static final int ANIM_INC   = 8;

  private Label movesLabel;
  private Label moveLabel;
  private List speedList;
  private List diskList;
  private Button startButton;
  private List methodList;
  private Vector[] ABC = new Vector[NUMSTACKS]; // init in diskSetup.
  private Graphics animG = null;

  public Hanoi() {
    int i, objY = 0, objX = 0, objWid = COMPWID;
    Label lbl;

    // calculate maxDiskNum (constrained by current width & height)
    maxDiskNum = (this.height - STACKTOP - COMPHT*2) / DISKHEIGHT; // !!!b
    int wnum = this.width / 24;
    maxDiskNum = (wnum < maxDiskNum) ? wnum-1 : maxDiskNum-1;
    if (maxDiskNum >= POW2.length)
      maxDiskNum = POW2.length-1;
    if (DEBUG) System.out.println("maxDiskNum: " + maxDiskNum);
    stackWidth = this.width / NUMSTACKS;
    stackBottom = STACKTOP + ((maxDiskNum+1) * DISKHEIGHT);
    if (DEBUG) System.out.println("stackWidth: "  + stackWidth);
    if (DEBUG) System.out.println("stackBottom: " + stackBottom);

    // TOP ROW: title, speed
    Title title = new Title("Tower of Hanoi");
    title.setRect(objX, objY, this.width, COMPHT);
    add(title);

    objWid = COMPWID+15;  // for list
    objX = this.width - (COMPWID + objWid + COMPSEP);
    lbl = new Label("Speed:");
    lbl.setRect(objX, objY, COMPWID, COMPHT);
    add(lbl);
    objX += COMPWID + COMPSEP;

    speedList = new List(SPEEDS);
    speedList.setRect(objX, objY, objWid, COMPHT);
    speedList.setSelectedIndex(SPEEDS.length - 1);
    add(speedList);

    // SECOND ROW: moves, move
    objX = 0;
    objY = COMPHT;
    lbl = new Label("Moves:");
    lbl.setRect(objX, objY, COMPWID, COMPHT);
    add(lbl);
    objX += COMPWID + COMPSEP;

    movesLabel = new Label("0");
    movesLabel.setRect(objX, objY, COMPWID, COMPHT);
    add(movesLabel);
    objX += COMPWID + COMPSEP;

    lbl = new Label("Move:");
    lbl.setRect(objX, objY, COMPWID, COMPHT);
    add(lbl);
    objX += COMPWID + COMPSEP;

    moveLabel = new Label("0");
    moveLabel.setRect(objX, objY, COMPWID, COMPHT);
    add(moveLabel);

    // just above bottom row !!!b
    lbl = new Label("weyer@kagi.com");
    lbl.setRect(0, this.height-2*COMPHT, 80, COMPHT);
    add(lbl);

    // BOTTOM ROW: start, disks, method
    objX = 0;
    objY = this.height - COMPHT;
    startButton = new Button("Start");
    startButton.setRect(objX, objY, COMPWID, COMPHT);
    add(startButton);
    objX += COMPWID + COMPSEP;

    lbl = new Label("Disks:");
    lbl.setRect(objX, objY, COMPWID, COMPHT);
    add(lbl);
    objX += COMPWID + COMPSEP;

    objWid = COMPWID-10; // give to next 2
    diskList = new List();
    for (i = 2; i <= maxDiskNum; i++)
      diskList.add(String.valueOf(i));
    diskList.setRect(objX, objY, objWid, COMPHT);
    diskList.setSelectedIndex(numDisks-2);
    add(diskList);
    objX += objWid + COMPSEP;

    objWid = COMPWID+5; // for label and list
    lbl = new Label("Method:");
    lbl.setRect(objX, objY, objWid, COMPHT);
    add(lbl);
    objX += objWid + COMPSEP;

    methodList = new List(METHODS);
    methodList.setRect(objX, objY, objWid, COMPHT);
    methodList.setSelectedIndex(moveMeth);
    add(methodList);

    diskSetup(numDisks);
  }

  public void onPaint(Graphics g) {
    if (DEBUG) System.out.println(
      "onPaint: " + ABC[0].getCount() + "," +
      ABC[1].getCount() + "," + ABC[2].getCount());
    Rect r;
    Vector v;
    int vlen;
    g.setColor(255,255,255);  // white
    g.fillRect(0, STACKTOP, this.width, stackBottom - STACKTOP); // clear
    g.setColor(0,0,0);        // black
    for (int col = 0; col < NUMSTACKS; col++) {
      v = ABC[col];
      vlen = v.getCount();
      for (int disk = 0; disk < vlen; disk++) {
        r = (Rect) v.get(disk);
        g.fillRect(r.x, r.y, r.width, DISKHEIGHT); //r.height);
      };
    };
    if (animG == null) { // !!!c. first time (doesn't work in MRJ if in constructor)
      animG = createGraphics();
      animG.setClip(0, STACKTOP, this.width, stackBottom - STACKTOP);
    };
  }

 /*
  private void flash() { // ?? for error
    animG.setDrawOp(Graphics.DRAW_XOR);
    animG.fillRect(0, STACKTOP, this.width, stackBottom - STACKTOP);
    Vm.sleep(500);
    animG.fillRect(0, STACKTOP, this.width, stackBottom - STACKTOP);
    animG.setDrawOp(Graphics.DRAW_OVER);
  }
*/

  private void diskSetup(int num) {
    if (DEBUG) System.out.println("diskSetup: " + num);
    for (int i = 1; i < NUMSTACKS; i++)
      ABC[i] = new Vector();
    ABC[0] = new Vector(num);
    int x = stackWidth / 2, w = x-2, xdiff = w / num, t = stackBottom-DISKHEIGHT;
    for (int i = 0; i < num; i++) {
      ABC[0].add(new Rect(x-w, t, w*2, DISKHEIGHT));
      w -= xdiff;
      t -= DISKHEIGHT;
    };
    moveNum = 0;
    movesLabel.setText(String.valueOf(POW2[num]-1));
    moveLabel.setText(String.valueOf(moveNum));
    repaint();
  }

  private void redrawDisk(int frX, int frY, int toX, int toY, int wid, int ht) {
    //stackG.setDrawOp(g.DRAW_OVER);
    animG.setColor(255,255,255);       // white
    animG.fillRect(frX, frY, wid, ht); // clear old
    animG.setColor(0,0,0);             // black
    animG.fillRect(toX, toY, wid, ht); // fill new
  }

  private void autoMove() {
    if (animArgs[ANIM_DIR] != ANIM_NONE) { // continue animation
      int x, y, d = DISKINC, wid = animArgs[ANIM_WID], ht = DISKHEIGHT; //animArgs[ANIM_HT];
      switch (animArgs[ANIM_DIR]) {
        case ANIM_UP:
 	  x = animArgs[ANIM_FRX];
          y = animArgs[ANIM_VAR];
          redrawDisk(x, y, x, y-d, wid, ht);
          y -= d;
          if (y > STACKTOP)
            animArgs[ANIM_VAR] = y; // continue up
          else {
            animArgs[ANIM_DIR] = ANIM_ACROSS;
            animArgs[ANIM_VAR] = x;
          };
          break;

        case ANIM_ACROSS:
          y = STACKTOP;
          x = animArgs[ANIM_VAR];
          d = animArgs[ANIM_INC];
          int toX = animArgs[ANIM_TOX];
          // if DISKINC != 1, make sure to stop
          if ((d > 1 && (x+d) > toX) || (d < -1 && (x+d) < toX))
            d = toX - x;
          redrawDisk(x, y, x+d, y, wid, ht);
          x += d;
          if (x != toX)
            animArgs[ANIM_VAR] = x;   // continue across
          else {
            animArgs[ANIM_DIR] = ANIM_DOWN;
            animArgs[ANIM_VAR] = y;
          };
          break;

        case ANIM_DOWN:
          x = animArgs[ANIM_TOX];
          y = animArgs[ANIM_VAR];
          redrawDisk(x, y, x, y+d, wid, ht);
          y += d;
          if (y < animArgs[ANIM_TOY])
            animArgs[ANIM_VAR] = y;           // continue down
          else {
            animArgs[ANIM_DIR] = ANIM_NONE;   // done
            moveLabel.setText(String.valueOf(moveNum));
	    //break;
          };
      };
    }

    else if (moveMeth == METHUSER || moveMeth == METHCHEAT || moveNum >= POW2[numDisks]-1)
      resetTimer(0);

    else {  // jump, move
      int k = moveNum+1, pk = 0, x = k;
      while ((x & 1) == 0) {
        pk++;
        x /= 2;
      };

      int gr = (((numDisks-pk) & 1) == 0) ? 1 : 2;
      x = k / POW2[pk];
      doMove((gr * ((x - 1) / 2)) % NUMSTACKS, (gr * ((x + 1) / 2)) % NUMSTACKS);
    };
  }

  private void doMove(int frPos, int toPos) {
    moveNum++;
    if (DEBUG) System.out.println("move: " + moveNum);

    Vector frVec = ABC[frPos];
    int frY = frVec.getCount() - 1;
    Rect rect = (Rect) frVec.get(frY);
    int rx = rect.x, ry = rect.y, rwid = rect.width, rht = DISKHEIGHT; //rect.height;
    frVec.del(frY);           // 'pop'

    Vector toVec = ABC[toPos];
    int toY = toVec.getCount();
    // 'translate' rect to new pos
    rect.x = rx + (toPos-frPos)*stackWidth;
    rect.y = ry + (frY-toY)*DISKHEIGHT;
    toVec.add(rect);  // 'push'

    if (moveMeth == METHJUMP) {
      moveLabel.setText(String.valueOf(moveNum));
      //if (DEBUG) System.out.println("Move Disk From " +
      //  STACKNAMES[frPos] + " To " + STACKNAMES[toPos]);
      redrawDisk(rx, ry, rect.x, rect.y, rwid, rht);
    }
    else {
      animArgs[ANIM_FRX]  = rx;
      animArgs[ANIM_VAR]  = ry; //animArgs[ANIM_FRY]  =
      animArgs[ANIM_TOX]  = rect.x;
      animArgs[ANIM_TOY]  = rect.y;
      animArgs[ANIM_WID]  = rwid;
      //animArgs[ANIM_HT]   = rht;
      animArgs[ANIM_DIR]  = ANIM_UP;
      animArgs[ANIM_INC]  = (toPos > frPos) ? DISKINC : -DISKINC;
    };
  }

  private void resetTimer(int ms) {
    if (appTimer != null)
      removeTimer(appTimer);
    if (ms > 0) {
      appTimer = addTimer(ms);
      startButton.setText("Stop");
    }
    else {
      appTimer = null;
      animArgs[ANIM_DIR] = ANIM_NONE;
      startButton.setText("Start");
    };
  }

  public void onEvent(Event ev) {
    Object obj = ev.target;
    int evType = ev.type;
    if (evType == ControlEvent.TIMER)
      autoMove();
    else if (obj == this && (evType == PenEvent.PEN_DOWN || evType == PenEvent.PEN_UP)) {
      //System.out.println("x: " + ((PenEvent) ev).x + "; y: " + ((PenEvent) ev).y);
      if (evType == PenEvent.PEN_DOWN)
        penX = ((PenEvent) ev).x;
      else { // PEN_UP
        int frPos = penX / stackWidth;
        // !!!b. ensure frPos&toPos are in range: 0 - NUMSTACKS-1
        if (frPos < 0)
          frPos = 0;
        else if (frPos >= NUMSTACKS)
          frPos = NUMSTACKS-1;
        int toPos = ((PenEvent) ev).x / stackWidth;
        if (toPos < 0)
          toPos = 0;
        else if (toPos >= NUMSTACKS)
          toPos = NUMSTACKS-1;
//System.out.println("x1: " + penX + "; x2: " + ((PenEvent) ev).x);
//System.out.println("frPos: " + frPos + "; toPos: " + toPos);
        Vector frVec = ABC[frPos];
        int frY = frVec.getCount();
        if (frPos == toPos || frY == 0)
          return; // same stack or empty source

        Vector toVec = ABC[toPos];
        int toY = toVec.getCount();
        if (moveMeth == METHCHEAT || toY == 0 ||
          ((Rect) frVec.get(frY-1)).width < ((Rect) toVec.get(toY-1)).width) {
          resetTimer(delay);
          doMove(frPos, toPos);
        };
        //else flash();
      };
    }
    else if (evType == ControlEvent.PRESSED) { // Button
      if (obj == startButton) {
//System.out.println(animArgs[ANIM_DIR]);
        //if (DEBUG) System.out.println("was: " + startButton.getText());
        if (moveMeth == METHUSER || moveMeth == METHCHEAT || appTimer == null) {
          diskSetup(numDisks);
          if (moveMeth > METHUSER && moveMeth < METHCHEAT)
            resetTimer(delay);
        }
        else
          resetTimer(0);
        //if (DEBUG) System.out.println("now: " + startButton.getText());
      }
      else if (obj == speedList) {
        if (DEBUG) System.out.println("Speed: " + speedList.getSelected());
        delay = DELAYS[speedList.getSelectedIndex()];
        resetTimer((appTimer == null) ? 0 : delay);
      }
      else if (obj == diskList) {
        if (DEBUG) System.out.println("Disks: " + diskList.getSelected());
        resetTimer(0);
        diskSetup(numDisks = diskList.getSelectedIndex() + 2);
      }
      else if (obj == methodList) {
        if (DEBUG) System.out.println("Method: " + methodList.getSelected());
        resetTimer(0);
        moveMeth = methodList.getSelectedIndex();
      }
      else if (DEBUG) System.out.println("ControlEvent? " + ev); // unexpected
    };
    //else if (DEBUG) System.out.println("Event? " + ev);
  }
}
